from formula import *
from encoders.printers import *


class BasicSmtUtil(object):
    
    def __init__(self):
        self.manager = VariableManager()
        self.assertions = []
        return

    def addAssertion(self, formula, simplify=True):
        if simplify:
            f = formula.simplify()
            self.assertions.append(f)
        else:
            self.assertions.append(formula)
        return

    pass



class LibSMT1Util(BasicSmtUtil):

    def __init__(self, logic="QF_RDL"):
        BasicSmtUtil.__init__(self)

        self.logic = logic
        return
    

    def printHeader(self, stream):
        stream.write("(benchmark TP\n")
        stream.write(" :source { FBK }\n")
        stream.write(" :logic %s" % self.logic)
        return

    
    def printFooter(self, stream):
        stream.write(")\n")
        return


    def dumpFalse(self, stream):
        stream.write(";; --> The encoder detected that this problem is trivially unsatisfiable <--\n\n")
        self.printHeader(stream)
        stream.write("\n")
        stream.write(" :formula (false)\n")
        self.printFooter(stream)
        return


    def dump(self, stream):
        self.printHeader(stream)

        stream.write("\n")

        for k in self.manager.variables.keys():
            if not k in self.manager.localVariables:
                v = self.manager.variables[k]
                assert not v is None

                if v.typename != Types.BOOL:
                    stream.write(" :extrafuns ((%s %s))\n" % (v.name, v.typename.toString(False)))
                else:
                    stream.write(" :extrapreds ((%s))\n" % v.name)

                stream.write("\n")


        stream.write(" :formula\n")

        p = SmtPrinter(stream, smt2version=False, strictSyntax=False)

        if len(self.assertions) > 1:
            stream.write("(and\n")
            for a in self.assertions:
                p.printer(a)
                stream.write("\n")
            stream.write(")\n")
        else:
            for a in self.assertions:
                p.printer(a)
                stream.write("\n")
        
        stream.write("\n")
        self.printFooter(stream)
        return

    pass



class LibSMT2Util(BasicSmtUtil):

    def __init__(self, model=False, logic="QF_RDL", proof=False):
        BasicSmtUtil.__init__(self)
        
        self.logic = logic
        self.proof = proof
        self.model = model
        
        return
    

    def printHeader(self, stream):
        stream.write("(set-option :print-success false)\n")
        if self.model:
            stream.write("(set-option :produce-models true)\n")
        if self.proof:
            stream.write("(set-option :produce-proofs true)\n")

        stream.write("(set-logic %s)\n" % self.logic)
        return


    def dumpFalse(self, stream):
        stream.write(";; --> The encoder detected that this problem is trivially unsatisfiable <--\n\n")
        self.printHeader(stream)
        stream.write("\n")
        stream.write("(assert false)\n")
        stream.write("\n")
        self.printFooter(stream)
        return


    def printFooter(self, stream):
        stream.write("(check-sat)\n")
        if self.model:
            stream.write("(get-model)\n")
        if self.proof:
            stream.write("(get-proof)\n")  
        stream.write("(exit)\n")
        return


    def printVar(self, stream, v):
        stream.write("(declare-fun %s %s)\n" % (v.name, v.typename.toString(True)))
        return


    def dump(self, stream, strict=True):
        self.printHeader(stream)

        stream.write("\n")

        for k in self.manager.variables.keys():
            if not k in self.manager.localVariables:
                v = self.manager.variables[k]
                assert not v is None
                self.printVar(stream, v)

        p = SmtPrinter(stream, strictSyntax=strict, smt2version=True)                

        stream.write("\n")

        for a in self.assertions:
            stream.write("(assert ")
            p.printer(a)
            stream.write(")\n")

        stream.write("\n")
        self.printFooter(stream)
        return

    pass
