from extractors.solvers.msat import *
from encoders.helpers import *
from encoders.printers import *

class StrategyChecker(object):

    def __init__(self, environment):
        self.environment = environment
        self.solver = MathSAT5Solver(environment)
        return


    def checkLinearStrategy(self, problem, strategy, clausifier):
        xvars = clausifier.xvars
        yvars = clausifier.yvars

        gamma = clausifier.getGamma()
        formula = clausifier.getPsi()

        for x in xvars:
            formula = Definition(x.name, strategy.getPolicyAsFormula(x), formula)

        formula = formula.simplify(True)

        self.solver.push()

        for y in yvars:
            self.solver.declareVariable(y)

        self.solver.addAssertion(gamma)
        self.solver.addAssertion(Not(formula))
        result = self.solver.solve()
        self.solver.pop()

        return (not result)


    def checkPiecewiseStrategy(self, problem, strategy, clausifier):
        xvars = clausifier.xvars
        yvars = clausifier.yvars

        gamma = clausifier.getGamma()
        psi = clausifier.getPsi()

        for y in yvars:
            self.solver.declareVariable(y)

        self.solver.push()

        self.solver.addAssertion(gamma)

        negation = []
        for (region, subpolicy) in strategy.getPieces():
            formula = psi
            for x in xvars:
                formula = Definition(x.name, subpolicy.getPolicyAsFormula(x), formula)

            # region -> not(formula)
            implication = Or(negation + [Not(region), Not(formula)])
            negation.append(region)
            self.solver.addAssertion(implication.simplify())

        result = self.solver.solve()

        # if result:
        #     printFormula(gamma)

        #     for y in yvars:
        #         v = self.solver.getModelValue(y)
        #         printFormula(Equals(y, Constant(v)))


        if result:
            print ("An infeasible point is:")
            for y in yvars:
                val = self.solver.getModelValue(y)
                print y.name, " = ", val

            print ("Checking strategy components")
            done = False
            for (region, subpolicy) in strategy.getPieces():
                val = self.solver.getModelValue(region)
                if val:
                    assert not done
                    done = True
                    print "The failing piece is:"
                    subpolicy.dump()

            if not done:
                print "The point is not covered by the strategy!"

        self.solver.pop()

        return (not result)

    pass
